/*******************************************************************************
 *
 * FILE: 	drivers/atm/iphase.c 
 *
 * DESCRIPTION: Interphase 5575 PCI ATM Device driver.
 *		Contains the main code for the device driver.
 *
 *
 * EDIT HISTORY:
 *
 * Monalisa Agrawal		12/98	Written the device driver
 * Deptt of Computer Science
 * Interoperability Lab
 * University of New Hampshire
 *
 * Peter Wang			5/99	Ported to linux kernel 2.2.5
 * Interphase Corp.
 * 
 ******************************************************************************/

#define MODULE  
#define DEBUG_TXPKT
#define DEBUG_RXPKT
  
#include <linux/module.h>  
#include <linux/config.h>  
#include <linux/kernel.h>  
#include <linux/mm.h>  
#include <linux/pci.h>  
#include <linux/errno.h>  
#include <linux/atm.h>  
#include <linux/atmdev.h>  
#include <linux/sonet.h>  
#include <linux/skbuff.h>  
#include <linux/time.h>  
#include <linux/sched.h> /* for xtime */  
#include <linux/delay.h>  
#include <linux/uio.h>  
#include <linux/init.h>  
#include <asm/system.h>  
#include <asm/io.h>  
#include <asm/uaccess.h>  
#include <asm/string.h>  
#include <asm/byteorder.h>  
#include <math.h>  
#include <linux/vmalloc.h>  
#include "iphase.h"		  
#include "suni.h"		  
#define IA_ABR 
#define swap(x) (((x & 0xff) << 8) | ((x & 0xff00) >> 8))  

static unsigned char ia_phy_get(struct atm_dev *dev, unsigned long addr);

#ifdef IA_ABR
// static ia_brd_regs_local_t  brd_regs_local;
#include "ia_abr.c"
static IADEV *ia_dev = NULL;
static struct timer_list ia_timer;
#endif

/* rest of the stuff is for debugging, copied from eni.c */  
  
#if 0  
#define DPRINTK(format,args...) printk(format,##args)  
#else  
#define DPRINTK(format,args...)  
#endif  
/*  
#ifdef IPHASE_DEBUG  
#ifdef TX_DEBUG  
#define TPRINTK(format,args...) printk(format,##args)  
#else   
#define TPRINTK(format,args...)  
#endif  
#ifdef RX_DEBUG  
#define RPRINTK(format,args...) printk(format,##args)  
#else   
#define RPRINTK(format,args...)  
#endif  
#endif  
*/  
  
  
#ifndef CONFIG_ATM_IPHASE5575_DEBUG  
  
  
#define NULLCHECK(x)  
#define EVENT(s,a,b)
/*  
void table_dump (IADEV *iadev, int status, 
                                                     int size, char *str) {
  u_short *reass_ptr = (u_short *)(iadev->reass_ram+REASS_TABLE);
  int i;
  printk(str);
  for (i = 0; i< size; i++)
   if (((i+1) % 16) == 0)
            printk(" %x\n", *(reass_ptr+i));
   else
            printk(" %x", *(reass_ptr+i));
}
 
static void event_dump(void)  
{  
}  
  
*/  
#else  
  
  
/*  
 * NULL pointer checking  
 */  
  
#define NULLCHECK(x) \
  if ((unsigned long) (x) < 0x30) printk(KERN_CRIT #x "==0x%x\n", (int) (x))  
  
/*  
 * Very extensive activity logging. Greatly improves bug detection speed but  
 * costs a few Mbps if enabled.  
 */  
  
#define EV 64  
  
static const char *ev[EV];  
static unsigned long ev_a[EV],ev_b[EV];  
static int ec = 0;  
  
  
static void EVENT(const char *s,unsigned long a,unsigned long b)  
{  
        ev[ec] = s;  
        ev_a[ec] = a;  
        ev_b[ec] = b;  
        ec = (ec+1) % EV;  
}  
  
static void event_dump(void)  
{  
        int n,i;  
  
        for (n = 0; n < EV; n++) {  
                i = (ec+n) % EV;  
                printk(KERN_NOTICE);  
                printk(ev[i] ? ev[i] : "(null)",ev_a[i],ev_b[i]);  
        }  
}  
  
  
#endif /* CONFIG_ATM_IPHASE5575_DEBUG */  
  
/* pwang_test debug utility */
int tcnter = 0, rcnter = 0;
void xdump( u_char*  cp, int  length, char*  prefix )
{
    int col, count;
    u_char prntBuf[120];
    u_char*  pBuf = prntBuf;
    count = 0;
    while(count < length){
        pBuf += sprintf( pBuf, "%s", prefix );
        for(col = 0;count + col < length && col < 16; col++){
            if (col != 0 && (col % 4) == 0)
                pBuf += sprintf( pBuf, " " );
            pBuf += sprintf( pBuf, "%02X ", cp[count + col] );
        }
        while(col++ < 16){      /* pad end of buffer with blanks */
            if ((col % 4) == 0)
                sprintf( pBuf, " " );
            pBuf += sprintf( pBuf, "   " );
        }
        pBuf += sprintf( pBuf, "  " );
        for(col = 0;count + col < length && col < 16; col++){
            if (isprint((int)cp[count + col]))
                pBuf += sprintf( pBuf, "%c", cp[count + col] );
            else
                pBuf += sprintf( pBuf, "." );
                }
        sprintf( pBuf, "\n" );
        // SPrint(prntBuf);
        printk(prntBuf);
        count += col;
        pBuf = prntBuf;
    }

}  /* close xdump(... */

  
static struct atm_dev *ia_boards = NULL;  
  
#define ACTUAL_RAM_BASE \
	RAM_BASE*((iadev->mem)/(128 * 1024))  
#define ACTUAL_SEG_RAM_BASE \
	IPHASE5575_FRAG_CONTROL_RAM_BASE*((iadev->mem)/(128 * 1024))  
#define ACTUAL_REASS_RAM_BASE \
	IPHASE5575_REASS_CONTROL_RAM_BASE*((iadev->mem)/(128 * 1024))  
  
  
/*-- some utilities and memory allocation stuff will come here -------------*/  
  
void desc_dbg(IADEV *iadev) {

  u_short tcq_wr_ptr, tcq_st_ptr, tcq_ed_ptr;
  u32 tmp, i;
  // regval = readl((u32)ia_cmds->maddr);
  tcq_wr_ptr =  readw(iadev->seg_reg+TCQ_WR_PTR);
  printk("B_tcq_wr = 0x%x desc = %d last desc = %d\n",
                     tcq_wr_ptr, readw(iadev->seg_ram+tcq_wr_ptr),
                     readw(iadev->seg_ram+tcq_wr_ptr-2));
  printk(" host_tcq_wr = 0x%x  host_tcq_rd = 0x%x \n",  iadev->host_tcq_wr, 
                   iadev->ffL.tcq_rd);
  tcq_st_ptr =  readw(iadev->seg_reg+TCQ_ST_ADR);
  tcq_ed_ptr =  readw(iadev->seg_reg+TCQ_ED_ADR);
  printk("tcq_st_ptr = 0x%x    tcq_ed_ptr = 0x%x \n", tcq_st_ptr, tcq_ed_ptr);
  i = 0;
  while (tcq_st_ptr != tcq_ed_ptr) {
      tmp = iadev->seg_ram+tcq_st_ptr;
      printk("TCQ slot %d desc = %d  Addr = 0x%x\n", i++, readw(tmp), tmp);
      tcq_st_ptr += 2;
  }
  for(i=0; i <TX_BUFFERS; i++)
      printk("Desc_tbl[%d] = %d \n", i, iadev->desc_tbl[i]);
} 
  
  
/*----------------------------- Recieving side stuff --------------------------*/  
 
static void rx_excp_rcvd(struct atm_dev *dev)  
{  
  IADEV *iadev;  
  u_short state;  
  u_short excpq_rd_ptr;  
  //u_short *ptr;  
  int vci, error;  
  
  iadev = IPHASE5575_DEV(dev);  
  
  state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
  while(!(state & EXCPQ_EMPTY))  
  {  
     excpq_rd_ptr = *(u16*)(iadev->reass_reg + EXCP_Q_RD_PTR);  
     if (excpq_rd_ptr != *(u16*)(iadev->reass_reg + EXCP_Q_WR_PTR))  
     {  
	/* we don't need to do anything right now.  
		later for keeping stats etc we might want to read   
		all this info.   
	*/  
	//ptr = *(u16 *)(iadev->reass_ram+excpq_rd_ptr);  
	vci = readw(iadev->reass_ram+excpq_rd_ptr);  
	error = readw(iadev->reass_ram+excpq_rd_ptr+2) & 0x0007;  
        // pwang_test
	excpq_rd_ptr += 4;  
	if (excpq_rd_ptr > *(u16*)(iadev->reass_reg + EXCP_Q_ED_ADR))  
		excpq_rd_ptr = *(u16*)(iadev->reass_reg + EXCP_Q_ST_ADR);  
	*(u16*)(iadev->reass_reg + EXCP_Q_RD_PTR) = excpq_rd_ptr;  
     }  
     /* bug found */  
     state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
  }  
}  
  
static void free_desc(struct atm_dev *dev, int desc)  
{  
	IADEV *iadev;  
	iadev = IPHASE5575_DEV(dev);  
        writew(desc, iadev->reass_ram+iadev->rfL.fdq_wr); 
	iadev->rfL.fdq_wr +=2;
	if (iadev->rfL.fdq_wr > iadev->rfL.fdq_ed)
		iadev->rfL.fdq_wr =  iadev->rfL.fdq_st;  
	writew(iadev->rfL.fdq_wr, iadev->reass_reg+FREEQ_WR_PTR);  
}  
  
  
static int rx_pkt(struct atm_dev *dev)  
{  
	IADEV *iadev;  
	struct atm_vcc *vcc;  
	unsigned short status;  
	struct rx_buf_desc *buf_desc_ptr;  
	int desc;   
	struct dle* wr_ptr;  
	int len;  
	struct sk_buff *skb;  
	u_int buf_addr, dma_addr;  
  
	iadev = IPHASE5575_DEV(dev);  
	if (iadev->rfL.pcq_rd == (readw(iadev->reass_reg+PCQ_WR_PTR)&0xffff)) 
	{  
   	    printk(KERN_ERR DEV_LABEL "(itf %d) Receive queue empty\n", dev->number);  
	    return -EINVAL;  
	}  
	/* mask 1st 3 bits to get the actual descno. */  
	desc = readw(iadev->reass_ram+iadev->rfL.pcq_rd) & 0x1fff;  
	/* update the read pointer  - maybe we shud do this in the end*/  
	if ( iadev->rfL.pcq_rd== iadev->rfL.pcq_ed) 
		iadev->rfL.pcq_rd = iadev->rfL.pcq_st;  
	else  
		iadev->rfL.pcq_rd += 2;
	writew(iadev->rfL.pcq_rd, iadev->reass_reg+PCQ_RD_PTR);  
  
  
	/* get the buffer desc entry.  
		update stuff. - doesn't seem to be any update necessary  
	*/  
	buf_desc_ptr = (struct rx_buf_desc *)(iadev->reass_ram+RX_DESC_BASE);  
	/* make the ptr point to the corresponding buffer desc entry */  
	buf_desc_ptr += desc;	  
	vcc = iadev->rx_open[buf_desc_ptr->vc_index & 0xffff];  
	if (!vcc)  
	{      
                free_desc(dev, desc); 
		//printk(KERN_NOTICE DEV_LABEL "(itf %d): Drop PDU !!!\n", 0);  
		return -1;  
	}  
	  
  
	/* might want to check the status bits for errors */  
	status = (u_short) (buf_desc_ptr->desc_mode);  
	if (status & (RX_CER | RX_PTE | RX_OFL))  
	{  
                vcc->stats->rx_err++;
		printk(KERN_ERR DEV_LABEL " bad packet, dropping it");  
                if (status & RX_CER)
                    printk(" cause: packet CRC error\n");
                else if (status & RX_PTE)
                    printk(" cause: packet time out\n");
                else 
                    printk(" cause: buffer over flow\n");
		free_desc(dev, desc);  
		return 0;  
	}  
  
	/*  
		build DLE.	  
		- I guess we allocate an skb and put the address in the   
			dle.  
		- this is not the hard part. The hard part is after   
		DMA is done. then how do u pass things up?  
	*/  
  
	buf_addr = (buf_desc_ptr->buf_start_hi << 16) | buf_desc_ptr->buf_start_lo;  
	dma_addr = (buf_desc_ptr->dma_start_hi << 16) | buf_desc_ptr->dma_start_lo;  
	len = dma_addr - buf_addr;  
		  
	/* lets allocate an skb for now */  
	skb = alloc_skb(len, GFP_ATOMIC);  
	if (!skb)  
	{  
		printk("can't allocate memory for recv\n");  
                vcc->stats->rx_drop++;
		return -1;  
	}  
	skb_put(skb,len);  
        // pwang_test
        ATM_SKB(skb)->vcc = vcc;
        ATM_DESC(skb) = desc;        
	skb_queue_tail(&iadev->rx_dma_q, skb);  

	/* Build the DLE structure */  
	wr_ptr = iadev->rx_dle_q.write;  
	wr_ptr->sys_pkt_addr = virt_to_bus(skb->data);	  
	wr_ptr->local_pkt_addr = buf_addr;  
	wr_ptr->bytes = len;	/* We don't know this do we ?? */  
	wr_ptr->mode = DMA_INT_ENABLE;  
  
	/* Put the descriptor number in the packet ready queue  
		and put the updated write pointer in the DLE field   
	- we don't have to do this. Instead, we have to do similar stuff  
		when DMA is complete, that is, put the desc in free queue.  
  
	- so what we do is put this desc in our own free q to be later put  
		in the SAR's rx free desc q when DMA is complete and we have   
		dequeued the packet.  
	*/  
  
	/* shud take care of wrap around here too. */  
        if(++wr_ptr == iadev->rx_dle_q.end)
             wr_ptr = iadev->rx_dle_q.start;
	iadev->rx_dle_q.write = wr_ptr;  
	udelay(1);  
	/* Increment transaction counter */  
	writel(1, iadev->dma+IPHASE5575_RX_COUNTER);   
	return 0;  
}  
  
static void rx_intr(struct atm_dev *dev)  
{  
  IADEV *iadev;  
  u_short status;  
  u_short state;  
  
  iadev = IPHASE5575_DEV(dev);  
  status = readl(iadev->reass_reg+REASS_INTR_STATUS_REG) & 0xffff;  
  if (status & RX_PKT_RCVD)  
  {  
	/* do something */  
	/* Basically recvd an interrupt for receving a packet.  
	A descriptor would have been written to the packet complete   
	queue. Get all the descriptors and set up dma to move the   
	packets till the packet complete queue is empty..  
	*/  
	state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
	while(!(state & PCQ_EMPTY))  
	{  
		rx_pkt(dev);  
		state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
	}  
  }  
  if (status & RX_FREEQ_EMPT)  
  {  
	state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
	if (state & FREEQ_EMPTY)  
	    writel(state | (RX_FREEQ_EMPT | RX_EXCP_RCVD | RX_PKT_RCVD), 
                                       iadev->reass_reg+REASS_MASK_REG);  
	    DPRINTK("Rx intr status: RX_FREEQ_EMPT %08x\n", status);  
  }  
  if (status & RX_EXCP_RCVD)  
  {  
	/* probably need to handle the exception queue also. */  
	// DPRINTK("Rx intr status: RX_EXCP_RCVD %08x\n", status);  
	rx_excp_rcvd(dev);  
  }  
  if (status & RX_RAW_RCVD)  
  {  
	/* need to handle the raw incoming cells. This deepnds on   
	whether we have programmed to receive the raw cells or not.  
	Else ignore. */  
	DPRINTK("Rx intr status:  RX_RAW_RCVD %08x\n", status);  
  }  
}  
  
  
static void rx_dle_intr(struct atm_dev *dev)  
{  
  IADEV *iadev;  
  struct atm_vcc *vcc;   
  struct sk_buff *skb;  
  int desc;  
  u_short state;   
  struct dle *dle, *cur_dle;  
  u_int dle_lp;  
  iadev = IPHASE5575_DEV(dev);  
 
  /* free all the dles done, that is just update our own dle read pointer   
	- do we really need to do this. Think not. */  
  /* DMA is done, just get all the recevie buffers from the rx dma queue  
	and push them up to the higher layer protocol. Also free the desc  
	associated with the buffer. */  
  dle = iadev->rx_dle_q.read;  
  dle_lp = readl(iadev->dma+IPHASE5575_RX_LIST_ADDR) & (sizeof(struct dle)*DLE_ENTRIES - 1);  
  cur_dle = (struct dle*)(iadev->rx_dle_q.start + (dle_lp >> 4));  
  while(dle != cur_dle)  
  {  
      /* free the DMAed skb */  
      skb = skb_dequeue(&iadev->rx_dma_q);  
      if (!skb)  
	 break;  
      desc = ATM_DESC(skb);
      free_desc(dev, desc);  
               
      if (!skb->len)  
      {  
          printk("rx_dle_intr: skb len 0\n");  
	  dev_kfree_skb(skb);  
      }  
      else  
      {  
          struct cpcs_trailer *trailer;
          u_short length;
          struct ia_vcc *ia_vcc;
          /* no VCC related housekeeping done as yet. lets see */  
          vcc = ATM_SKB(skb)->vcc;
	  if (!vcc) {
	      printk(KERN_CRIT DEV_LABEL "(itf %d): INVALID SKB !!!\n", 0);  
               vcc->stats->rx_err++;
              dev_kfree_skb(skb);
              return;
          }
          ia_vcc = IPHASE5575_VCC(vcc);
          if (ia_vcc == NULL)
          {
             vcc->stats->rx_err++;
             dev_kfree_skb(skb);
             return;
           }
          // get real pkt length  pwang_test
          trailer = (struct cpcs_trailer*)((u_char *)skb->data +
                                 skb->len - sizeof(struct cpcs_trailer));
          length =  swap(trailer->length);
          if ((length > 8192) || (length > 
                              (skb->len - sizeof(struct cpcs_trailer))))
          {
             vcc->stats->rx_err++;
             dev_kfree_skb(skb);
             printk("rx_dle_intr: Bad len in AAL5 trailer %d (skb len %d)", 
                                                            length, skb->len);
             return; 
          }
          skb_trim(skb, length);
          
#ifdef DEBUG_RXPKT                       
	  /* Display the packet */  
	  //printk("\nDmad Recvd data: len = %d \n", skb->len);  
          //xdump(skb->data, skb->len, "RX: ");
          //printk("\n");
#endif /* DEBUG_TXPKT */
	  DPRINTK("rx_dle_intr: skb push");  
	  vcc->push(vcc,skb);  
	  DPRINTK("rx_dle_intr: skb free");  
	  vcc->stats->rx++;   
      }  
      if (++dle == iadev->rx_dle_q.end)  
    	  dle = iadev->rx_dle_q.start;  
  }  
  iadev->rx_dle_q.read = dle;  
  
  /* if the interrupts are masked because there were no free desc available,  
		unmask them now. */  
  state = readl(iadev->reass_reg + STATE_REG) & 0xffff;  
  if (!(state & FREEQ_EMPTY))  
     writel(state & ~(RX_FREEQ_EMPT | RX_EXCP_RCVD | RX_PKT_RCVD), 
                                      iadev->reass_reg+REASS_MASK_REG);  
  DPRINTK("rx_dle_intr: done\n");  
}  
  
  
  
static int open_rx(struct atm_vcc *vcc)  
{  
	IADEV *iadev;  
	u_short *vc_table;  
	u_short *reass_ptr;  
	DPRINTK("iadev: open_rx %d.%d\n", vcc->vpi, vcc->vci);

	if (vcc->qos.rxtp.traffic_class == ATM_NONE) return 0;    
	iadev = IPHASE5575_DEV(vcc->dev);  
	/* Make only this VCI in the vc table valid and let all   
		others be invalid entries */  
	vc_table = (u_short *)(iadev->reass_ram+RX_VC_TABLE);  
	vc_table += vcc->vci;  
	/* mask the last 6 bits and OR it with 3 for 1K VCs */  
	*vc_table = (*vc_table & 0xffc0) | 3;  
  
	/* Also keep a list of open rx vcs so that we can attach them with  
		incoming PDUs later. */  
	if ((vcc->qos.rxtp.traffic_class == ATM_ABR) || (vcc->qos.txtp.traffic_class == ATM_ABR))  
	{       srv_cls_param_t srv_p;
printk("Rx ABR VCC -------------------------\n");
                init_abr_vc(&srv_p);
                ia_open_abr_vc(iadev, &srv_p, vcc, 0);
	} 
       	else {  /* for UBR  later may need to add CBR logic */
        	reass_ptr = (u_short *)(iadev->reass_ram+REASS_TABLE);
           	reass_ptr += vcc->vci;  
           	*reass_ptr = NO_AAL5_PKT;
       	}
	
	if (iadev->rx_open[vcc->vci])  
		printk(KERN_CRIT DEV_LABEL "(itf %d): VCI %d already open\n",  
			vcc->dev->number, vcc->vci);  
	iadev->rx_open[vcc->vci] = vcc;  
	return 0;  
}  
  
static int rx_init(struct atm_dev *dev)  
{  
	IADEV *iadev;  
	struct rx_buf_desc *buf_desc_ptr;  
	unsigned long rx_pkt_start;  
	u32 *dle_addr;  
	struct abr_vc_table  *abr_vc_table; 
	u16 *vc_table;  
	u16 *reass_table;  
                   u16 *ptr16;
	int i;  
	u_short freeq_st_adr;  
	u_short *freeq_start;  
  
	iadev = IPHASE5575_DEV(dev);  
  
	/* I need to initialize the DLEs somewhere. Lets see what I   
		need to do for this, hmmm...  
		- allocate memory for 256 DLEs. make sure that it starts  
		on a 4k byte address boundary. Program the start address   
		in Receive List address register.  ..... to do for TX also  
	   To make sure that it is a 4k byte boundary - allocate 8k and find   
		4k byte boundary within.  
		( (addr + (4k-1)) & ~(4k-1) )  
	*/   
  
	/* allocate 8k bytes */  
	dle_addr = (u32*)kmalloc(2*sizeof(struct dle)*DLE_ENTRIES, GFP_KERNEL);  
	if (!dle_addr)  
	{  
		printk(KERN_ERR DEV_LABEL "can't allocate DLEs\n");  
	}  
	/* find 4k byte boundary within the 8k allocated */  
	dle_addr = (u32*)( ((u32)dle_addr+(4096-1)) & ~(4096-1) );  
	iadev->rx_dle_q.start = (struct dle*)dle_addr;  
	iadev->rx_dle_q.read = iadev->rx_dle_q.start;  
	iadev->rx_dle_q.write = iadev->rx_dle_q.start;  
	iadev->rx_dle_q.end = (struct dle*)((u32)dle_addr+sizeof(struct dle)*DLE_ENTRIES);  
	/* the end of the dle q points to the entry after the last  
	DLE that can be used. */  
  
	/* write the upper 20 bits of the start address to rx list address register */  
	writel(virt_to_bus(dle_addr) & 0xfffff000, iadev->dma+IPHASE5575_RX_LIST_ADDR);  
	DPRINTK("Tx Dle list addr: 0x%08x value: 0x%0x\n", (u32)(iadev->dma+IPHASE5575_TX_LIST_ADDR), *(u32*)(iadev->dma+IPHASE5575_TX_LIST_ADDR));  
	DPRINTK("Rx Dle list addr: 0x%08x value: 0x%0x\n", (u32)(iadev->dma+IPHASE5575_RX_LIST_ADDR), *(u32*)(iadev->dma+IPHASE5575_RX_LIST_ADDR));  
  
	writew(0xffff, iadev->reass_reg+REASS_MASK_REG);  
	writew(0, iadev->reass_reg+MODE_REG);  
	writew(RESET_REASS, iadev->reass_reg+REASS_COMMAND_REG);  
  
	/* Receive side control memory map  
	   -------------------------------  
  
		Buffer descr	0x0000 (736 - 23K)  
		VP Table	0x5c00 (256 - 512)  
		Except q	0x5e00 (128 - 512)  
		Free buffer q	0x6000 (1K - 2K)  
		Packet comp q	0x6800 (1K - 2K)  
		Reass Table	0x7000 (1K - 2K)  
		VC Table	0x7800 (1K - 2K)  
		ABR VC Table	0x8000 (1K - 32K)  
	*/  
	  
	/* need to initialize the buffer descriptor table   
		- now our table has 736 entries. So descriptor numbers  
		range from 1 to 735(0 unused). Initialize each entry to   
		basically point to the associated packet buffer location.   
		Might need to intialize a few other fields.  
	 Two things to be done.  
	 1. Decide on the way packet memory is to be divided. Accordingly,   
		initialize the buffer descriptor table for each entry to point  
		to the start of the associated packet buffer.  
	 2. Fill the free buffer queue in the reassembly RAM with all the valid  
		descriptor numbers.  
	*/  
	  
	/* Base address for Buffer Descriptor Table */  
	writew(0x0000, iadev->reass_reg+REASS_DESC_BASE);  
	/* Set the buffer size register */  
	writew(RX_BUF_SZ, iadev->reass_reg+BUF_SIZE);  
  
	/* Initialize each entry in the Buffer Descriptor Table */  
	buf_desc_ptr =(struct rx_buf_desc *)(iadev->reass_ram+RX_DESC_BASE);  
	memset((caddr_t)buf_desc_ptr, 0, sizeof(struct rx_buf_desc));  
	buf_desc_ptr++;  
	rx_pkt_start = RX_PACKET_RAM;  
	for(i=1; i<=RX_BUFFERS; i++)  
	{  
		memset((caddr_t)buf_desc_ptr, 0, sizeof(struct rx_buf_desc));  
		buf_desc_ptr->buf_start_hi = rx_pkt_start >> 16;  
		buf_desc_ptr->buf_start_lo = rx_pkt_start & 0x0000ffff;  
		buf_desc_ptr++;		  
		rx_pkt_start += RX_BUF_SZ;  
	}  
	DPRINTK("Rx Buffer desc ptr: 0x%0x\n", (u32)(buf_desc_ptr));  
		    
	/* Communication Queues base address */  
	writew(0x0000, iadev->reass_reg+REASS_QUEUE_BASE);  
  
	/* Free Buffer Descriptor Queue - wrongly referred in many places  
			as Free Buffer Queue */  
	writew(0x6000, iadev->reass_reg+FREEQ_ST_ADR);  
	writew(0x67fe, iadev->reass_reg+FREEQ_ED_ADR);  
	writew(0x6000, iadev->reass_reg+FREEQ_RD_PTR);  
	writew(0x6000, iadev->reass_reg+FREEQ_WR_PTR);  
        
	/* Fill the FREEQ with all the free descriptors. */  
	freeq_st_adr = readw(iadev->reass_reg+FREEQ_ST_ADR);  
	freeq_start = (u_short *)(iadev->reass_ram+freeq_st_adr);  
	for(i=1; i<=RX_BUFFERS; i++)  
	{  
		*freeq_start = (u_short)i;  
		freeq_start++;  
	}  
	DPRINTK("freeq_start: 0x%0x\n", (u32)freeq_start);  
	/* update the write pointer */  
	writew(0x6000+RX_BUFFERS*sizeof(u_short), iadev->reass_reg+FREEQ_WR_PTR);  
	writew(0x6000+RX_BUFFERS*sizeof(u_short), iadev->reass_reg+FREEQ_ED_ADR);  
  
	/* Packet Complete Queue */  
	writew(0x6800, iadev->reass_reg+PCQ_ST_ADR);  
	writew(0x6ffe, iadev->reass_reg+PCQ_ED_ADR);  
	writew(0x6800, iadev->reass_reg+PCQ_RD_PTR);  
	writew(0x6800, iadev->reass_reg+PCQ_WR_PTR);  
  
	/* Exception Queue */  
	writew(0x5e00, iadev->reass_reg+EXCP_Q_ST_ADR);  
	writew(0x5ffc, iadev->reass_reg+EXCP_Q_ED_ADR);  
	writew(0x5e00, iadev->reass_reg+EXCP_Q_RD_PTR);  
	writew(0x5e00, iadev->reass_reg+EXCP_Q_WR_PTR);  

    	/* Load local copy of FREEQ and PCQ ptrs */
        iadev->rfL.fdq_st = readw(iadev->reass_reg+FREEQ_ST_ADR) & 0xffff;
       	iadev->rfL.fdq_ed = readw(iadev->reass_reg+FREEQ_ED_ADR) & 0xffff ;
	iadev->rfL.fdq_rd = readw(iadev->reass_reg+FREEQ_RD_PTR) & 0xffff;
	iadev->rfL.fdq_wr = readw(iadev->reass_reg+FREEQ_WR_PTR) & 0xffff;
        iadev->rfL.pcq_st = readw(iadev->reass_reg+PCQ_ST_ADR) & 0xffff;
	iadev->rfL.pcq_ed = readw(iadev->reass_reg+PCQ_ED_ADR) & 0xffff;
	iadev->rfL.pcq_rd = readw(iadev->reass_reg+PCQ_RD_PTR) & 0xffff;
	iadev->rfL.pcq_wr = readw(iadev->reass_reg+PCQ_WR_PTR) & 0xffff;
	
		  
	/* just for check - no VP TBL */  
	/* VP Table */  
	/* writew(0x0b80, iadev->reass_reg+VP_LKUP_BASE); */  
	/* initialize VP Table for invalid VPIs  
		- I guess we can write all 1s or 0x000f in the entire memory  
		  space or something similar.  
	*/  
  
	/* This seems to work and looks right to me too !!! */  
	writew(0x0e00, iadev->reass_reg+REASS_TABLE_BASE);   
	/* initialize Reassembly table to I don't know what ???? */  
	reass_table = (u16 *)(iadev->reass_ram+REASS_TABLE);  
	for(i=0; i<REASS_TABLE_SZ; i++)  
		*reass_table = NO_AAL5_PKT;  
	/* VC Table */  
	writew(0x0f03, iadev->reass_reg+VC_LKUP_BASE);  
	/* initialize VC Table for invalid VCIs  
		- I guess we can write all 1s or 0x000f in the entire memory  
		  space or something similar.  
		- or we can initialize each VC entry to point to the right  
		  reassembly table entry. - chosing this approachwe reduce   
		  some headache at VC setup time.  
	*/  
	vc_table = (u16 *)(iadev->reass_ram+RX_VC_TABLE);  
	for(i=0; i<RX_VC_TABLE_SZ; i++)  
	{  
		/* shift the reassembly pointer by 3 + lower 3 bits of   
		vc_lkup_base register (=3 for 1K VCs) and the last byte   
		is those low 3 bits.   
		Shall program this later.  
		*/  
		*vc_table = (i << 6) | 15;	/* for invalid VCI */  
		vc_table++;  
	}  
  
	/* ABR VC Table */  
	writew(0x1000, iadev->reass_reg+ABR_LKUP_BASE);                     
	abr_vc_table = (struct abr_vc_table *)(iadev->reass_ram+ABR_VC_TABLE);  
        memset ((char*)abr_vc_table,0,REASS_TABLE_SZ*sizeof(struct abr_vc_table ) );
    	for(i=0; i<REASS_TABLE_SZ; i++)  	{   		
		abr_vc_table->rdf = 0x0003;
             		abr_vc_table->air = 0x5eb1;
	       	abr_vc_table++;   	
                   }  

	/* Initialize other registers */  
  
	/* VP Filter Register set for VC Reassembly only */  
	writew(0xff00, iadev->reass_reg+VP_FILTER);  

           	writew(0, iadev->reass_reg+XTRA_RM_OFFSET);
	writew(0x1,  iadev->reass_reg+PROTOCOL_ID);

	/*   Packet Timeout Count  related Registers :         pwang_test
	        Set packet timeout to occur in about 3 seconds
	        Set Packet Aging Interval count register to overflow in about   4 us
 	*/  
        writew(0xF6F8, iadev->reass_reg+PKT_TM_CNT );
        ptr16 = (u16*)(iadev->reass_reg+REASS_TABLE);
        i = ((u32)ptr16 >> 6) & 0xff;
	ptr16  += REASS_TABLE_SZ -1;
	i |=(((u32)ptr16 << 2) & 0xff00);
        writew(i, iadev->reass_reg+TMOUT_RANGE);

        /* initiate the desc_tble */
        for(i=0; i<TX_BUFFERS;i++)
            iadev->desc_tbl[i] = 0;

	/* Mode Register */  
	writew(R_ONLINE, iadev->reass_reg+MODE_REG);  
  
	/* to clear the interrupt status register - read it */  
	readw(iadev->reass_reg+REASS_INTR_STATUS_REG);   
  
	/* Mask Register - clear it */  
	writew(0, iadev->reass_reg+REASS_MASK_REG);  
  
	skb_queue_head_init(&iadev->rx_dma_q);  
	iadev->rx_free_desc_qhead = NULL;   
	iadev->rx_open = (struct atm_vcc **)get_free_page(GFP_KERNEL);  
	if (!iadev->rx_open)  
	{  
		printk(KERN_ERR DEV_LABEL "(itf %d): couldn't get free page\n",  
		dev->number);  
		return -ENOMEM;  
	}  
	memset(iadev->rx_open, 0, PAGE_SIZE);  
	return 0;  
}  
  
  
/*------------------------------ Transmit side stuff --------------------------*/  
/*  
	The memory map suggested in appendix A and the coding for it.   
	Keeping it around just in case we change our mind later.  
  
		Buffer descr	0x0000 (128 - 4K)  
		UBR sched	0x1000 (1K - 4K)  
		UBR Wait q	0x2000 (1K - 4K)  
		Commn queues	0x3000 Packet Ready, Trasmit comp(0x3100)  
					(128 - 256) each  
		extended VC	0x4000 (1K - 8K)  
		ABR sched	0x6000	and ABR wait queue (1K - 2K) each  
		CBR sched	0x7000 (as needed)  
		VC table	0x8000 (1K - 32K)  
  
	writew(0x0000, iadev->seg_reg+SEG_DESC_BASE);  
	writew(0x0000, iadev->seg_reg+SEG_QUEUE_BASE);  
  
	writew(0x3100, iadev->seg_reg+TCQ_ST_ADR);  
	writew(0x31fe, iadev->seg_reg+TCQ_ED_ADR);  
	writew(0x3100, iadev->seg_reg+TCQ_RD_PTR);  
	writew(0x3100, iadev->seg_reg+TCQ_WR_PTR);  
  
	writew(0x3000, iadev->seg_reg+PRQ_ST_ADR);  
	writew(0x30fe, iadev->seg_reg+PRQ_ED_ADR);  
	writew(0x3000, iadev->seg_reg+PRQ_RD_PTR);  
	writew(0x3000, iadev->seg_reg+PRQ_WR_PTR);  
  
	writew(0x0002, iadev->seg_reg+UBR_SBPTR_BASE);  
	writew(0x0040, iadev->seg_reg+UBRWQ_BASE);  
  
	writew(0x000c, iadev->seg_reg+ABR_SBPTR_BASE);  
	writew(0x00d0, iadev->seg_reg+ABRWQ_BASE);  
  
	writew(0x0083, iadev->seg_reg+VCT_BASE);  
	writew(0x0040, iadev->seg_reg+VCTE_BASE);  
*/  
  
  
static void tx_intr(struct atm_dev *dev)  
{  
	IADEV *iadev;  
	/* struct free_desc_q *entry; */  
	/* int desc; */  
	unsigned short status;  
	/* unsigned short tcq_rd_ptr; */  
	/* u_short intr_mask; */  
	static int entered = 0;  
  
	entered++;  
	iadev = IPHASE5575_DEV(dev);  
  
	status = readl(iadev->seg_reg+SEG_INTR_STATUS_REG);  
	  
	if (status & TCQ_NOT_EMPTY)  
	{  
	    DPRINTK("Tx intr done\n");  
	}  
	DPRINTK("Tx intr done\n");  
}  
  
static void tx_dle_intr(struct atm_dev *dev)
{
        IADEV *iadev;
        iadev = IPHASE5575_DEV(dev);
	DPRINTK("Tx dle intr called\n");  
}
  
static int open_tx(struct atm_vcc *vcc)  
{  
	struct ia_vcc *ia_vcc;  
	IADEV *iadev;  
	struct main_vc *vc;  
	struct ext_vc *evc;  
  
	DPRINTK("iadev: open_tx entered\n");  
	ia_vcc = IPHASE5575_VCC(vcc);  
	iadev = IPHASE5575_DEV(vcc->dev);  
  
	if (vcc->qos.txtp.traffic_class == ATM_NONE) return 0;  
	  
	vc = (struct main_vc *)(iadev->seg_ram+MAIN_VC_TABLE);  
	evc = (struct ext_vc *)(iadev->seg_ram+EXT_VC_TABLE);  
	vc += vcc->vci;  
	evc += vcc->vci;  
	memset((caddr_t)vc, 0, sizeof(struct main_vc));  
	memset((caddr_t)evc, 0, sizeof(struct ext_vc));  
	  
	/* store the most significant 4 bits of vci as the last 4 bits   
		of first part of atm header.  
	   store the last 12 bits of vci as first 12 bits of the second  
		part of the atm header.  
	*/  
	evc->atm_hdr1 = (vcc->vci >> 12) & 0x000f;  
	evc->atm_hdr2 = (vcc->vci & 0x0fff) << 4;  
  
	/* check the following for different traffic classes */  
	if (vcc->qos.txtp.traffic_class == ATM_UBR)  
	{  
		vc->type = UBR;  
		vc->status = CRC_APPEND;
                if (vcc->qos.txtp.max_pcr > 0) {
                   vc->acr = cellrate_to_float(vcc->qos.txtp.max_pcr);  
                }
		else vc->acr = RATE155;  
// printk("UBR: vcc->qos.txtp.max_pcr = %d\n", vc->acr);
	}  
	else if (vcc->qos.txtp.traffic_class == ATM_ABR)  
	{       srv_cls_param_t srv_p;
		DPRINTK("Tx ABR VCC\n");  
                init_abr_vc(&srv_p);
                if (vcc->qos.txtp.max_pcr > 0)
                    srv_p.pcr = vcc->qos.txtp.max_pcr;
//printk("vcc->qos.txtp.max_pcr = %d \n", srv_p.pcr);
		ia_open_abr_vc(iadev, &srv_p, vcc, 1);
	}  
	else  
		DPRINTK("iadev: traffic classes other than UBR and ABR not  
			handled currently\n");  
	DPRINTK("ia open_tx returning \n");  
	return 0;  
}  
  
  
static int tx_init(struct atm_dev *dev)  
{  
	IADEV *iadev;  
	struct tx_buf_desc *buf_desc_ptr;
	unsigned int tx_pkt_start;  
	u32 *dle_addr;  
	int i;  
	u_short tcq_st_adr;  
	u_short *tcq_start;  
	u_short prq_st_adr;  
	u_short *prq_start;  
	struct main_vc *vc;  
	struct ext_vc *evc;   
  
	iadev = IPHASE5575_DEV(dev);  
	DPRINTK("Tx MASK REG: 0x%0x\n", readw(iadev->seg_reg+SEG_MASK_REG));  
  
	/*---------- Initializing Transmit DLEs ----------*/  
	/* allocating 8k memory for transmit DLEs */  
	dle_addr = (u32*)kmalloc(2*sizeof(struct dle)*DLE_ENTRIES, GFP_KERNEL);  
	if (!dle_addr)  
	{  
		printk(KERN_ERR DEV_LABEL "can't allocate TX DLEs\n");  
	}  
  
	/* find 4k byte boundary within the 8k allocated */  
	dle_addr = (u32*)(((u32)dle_addr+(4096-1)) & ~(4096-1));  
	iadev->tx_dle_q.start = (struct dle*)dle_addr;  
	iadev->tx_dle_q.read = iadev->tx_dle_q.start;  
	iadev->tx_dle_q.write = iadev->tx_dle_q.start;  
	iadev->tx_dle_q.end = (struct dle*)((u32)dle_addr+sizeof(struct dle)*DLE_ENTRIES);  

	/* So the end points to the structure after the last DLE   
	that can be used. */  
  
	/* write the upper 20 bits of the start address to tx list address register */  
	writel(virt_to_bus(dle_addr) & 0xfffff000, iadev->dma+IPHASE5575_TX_LIST_ADDR);  
	writew(0xffff, iadev->seg_reg+SEG_MASK_REG);  
	writew(0, iadev->seg_reg+MODE_REG_0);  
	writew(RESET_SEG, iadev->seg_reg+SEG_COMMAND_REG);  
  
	/*  
	   currently - 1023 VC operation control memory map chosen from appendix A  
  
	   Transmit side control memory map  
	   --------------------------------  
	  
	Alternate memory map suggested in 5575 PCI ATM Driver Porting Guide  
	page 7 as below. Currently, this is programmed in the driver.  
  
	 	Buffer descr 	0x0000 (128 - 4K)  
		Commn queues	0x1000	Transmit comp, Packet ready(0x1400)   
					(512 - 1K) each  
					TCQ - 4K, PRQ - 5K  
		CBR Table 	0x1800 (as needed) - 6K  
		UBR Table	0x3000 (1K - 4K) - 12K  
		UBR Wait queue	0x4000 (1K - 4K) - 16K  
		ABR sched	0x5000	and ABR wait queue (1K - 2K) each  
				ABR Tbl - 20K, ABR Wq - 22K   
		extended VC	0x6000 (1K - 8K) - 24K  
		VC Table	0x8000 (1K - 32K) - 32K  
	  
	Between 0x2000 (8K) and 0x3000 (12K) there is 4K space left for VBR Tbl  
	and Wait q, which can be allotted later.  
	*/  
  
	/* need to initialize the buffer descriptor table   
		- now our table has 128 entries. So descriptor numbers  
		range from 1 to 127(0 unused). Initialize each entry to   
		basically point to the associated packet buffer location.   
		Might need to intialize a few other fields.  
	 Two things to be done.  
	 1. Decide on the way packet memory is to be divided. Accordingly,   
		initialize the buffer descriptor table for each entry to point  
		to the start of the associated packet buffer.  
	 2. Build a free buffer queue. Build it either as a table or a queue.  
		table - init is easy. each lookup takes time.  
		queue - init takes time. but lookups are easy.  
	*/  
  
	/* Buffer Descriptor Table Base address */  
	writew(0x0000, iadev->seg_reg+SEG_DESC_BASE);  
  
	/* initialize each entry in the buffer descriptor table */  
	buf_desc_ptr =(struct tx_buf_desc *)(iadev->seg_ram+TX_DESC_BASE);  
	memset((caddr_t)buf_desc_ptr, 0, sizeof(struct tx_buf_desc));  
	buf_desc_ptr++;  
	tx_pkt_start = TX_PACKET_RAM;  
	for(i=1; i<=TX_BUFFERS; i++)  
	{  
		memset((caddr_t)buf_desc_ptr, 0, sizeof(struct tx_buf_desc));  
		buf_desc_ptr->desc_mode = AAL5;  
		buf_desc_ptr->buf_start_hi = tx_pkt_start >> 16;  
		buf_desc_ptr->buf_start_lo = tx_pkt_start & 0x0000ffff;  
		buf_desc_ptr++;		  
		tx_pkt_start += TX_BUF_SZ;  
	}  
       	for (i= 0; i< TX_BUFFERS; i++)
       	{
       	    iadev->tx_buf[i] =(caddr_t)kmalloc(TX_BUF_SZ,GFP_KERNEL|GFP_DMA);
            if(!iadev->tx_buf[i]) {                
		printk(KERN_ERR DEV_LABEL " couldn't get freepage\n");            
         	return -EAGAIN;
            }
        }
  
	/* Communication Queues base address */  
	writew(0x0000, iadev->seg_reg+SEG_QUEUE_BASE);  
  
	/* Transmit Complete Queue */  
	writew(0x1000, iadev->seg_reg+TCQ_ST_ADR);  
	writew(0x1000, iadev->seg_reg+TCQ_RD_PTR);  

	/* Fill the TCQ with all the free descriptors. */  
	tcq_st_adr = readw(iadev->seg_reg+TCQ_ST_ADR);  
	tcq_start = (u_short *)(iadev->seg_ram+tcq_st_adr);  
	for(i=1; i<=TX_BUFFERS; i++)  
	{  
		*tcq_start = (u_short)i;  
		tcq_start++;  
	}  
	/* update the write pointer */  
	writew(0x1000+TX_BUFFERS*sizeof(u_short), iadev->seg_reg+TCQ_WR_PTR); 

	iadev->host_tcq_wr = 0x1000+TX_BUFFERS*sizeof(u_short);
        writew(0x1000+2*TX_BUFFERS*sizeof(u_short), iadev->seg_reg+TCQ_ED_ADR); 
  
	/* Packet Ready Queue */  
	writew(0x1400, iadev->seg_reg+PRQ_ST_ADR);  
	writew(0x1400+2*TX_BUFFERS*sizeof(u_short), iadev->seg_reg+PRQ_ED_ADR);
	writew(0x1400, iadev->seg_reg+PRQ_RD_PTR);  
	writew(0x1400, iadev->seg_reg+PRQ_WR_PTR);  
	 
        /* Load local copy of PRQ and TCQ ptrs */
        iadev->ffL.prq_st = readw(iadev->seg_reg+PRQ_ST_ADR) & 0xffff;
	iadev->ffL.prq_ed = readw(iadev->seg_reg+PRQ_ED_ADR) & 0xffff;
 	iadev->ffL.prq_wr = readw(iadev->seg_reg+PRQ_WR_PTR) & 0xffff;

	iadev->ffL.tcq_st = readw(iadev->seg_reg+TCQ_ST_ADR) & 0xffff;
	iadev->ffL.tcq_ed = readw(iadev->seg_reg+TCQ_ED_ADR) & 0xffff;
	iadev->ffL.tcq_rd = readw(iadev->seg_reg+TCQ_RD_PTR) & 0xffff;

	/* Just for safety initializing the queue to have desc 1 always */  
	/* Fill the PRQ with all the free descriptors. */  
	prq_st_adr = readw(iadev->seg_reg+PRQ_ST_ADR);  
	prq_start = (u_short *)(iadev->seg_ram+prq_st_adr);  
	for(i=1; i<=TX_BUFFERS; i++)  
	{  
		*prq_start = (u_short)0;	/* desc 1 in all entries */  
		prq_start++;  
	}  
  
	/* CBR Table */  
  
	/* UBR scheduling Table and wait queue */  
	/* initialize all bytes of UBR scheduler table and wait queue to 0   
		- SCHEDSZ is 1K (# of entries).  
		- UBR Table size is 4K  
		- UBR wait queue is 4K  
	   since the table and wait queues are contiguous, all the bytes   
	   can be intialized by one memeset.  
	*/  
	writew(0x0006, iadev->seg_reg+UBR_SBPTR_BASE);  
	writew(0x0080, iadev->seg_reg+UBRWQ_BASE);  
	memset((caddr_t)(iadev->seg_ram+UBR_SCHED_TABLE), 0, SCHEDSZ*8);   
  
	/* ABR scheduling Table(0x5000-0x57ff) and wait queue(0x5800-0x5fff)*/  
	/* initialize all bytes of ABR scheduler table and wait queue to 0   
		- SCHEDSZ is 1K (# of entries).  
		- ABR Table size is 2K  
		- ABR wait queue is 2K  
	   since the table and wait queues are contiguous, all the bytes   
	   can be intialized by one memeset.  
	*/  
	writew(0x000a, iadev->seg_reg+ABR_SBPTR_BASE);  
	writew(0x00b0, iadev->seg_reg+ABRWQ_BASE);  
	memset((caddr_t)(iadev->seg_ram+ABR_SCHED_TABLE), 0, SCHEDSZ*4);  
  
	/* Main and Extended VC Table */  
	writew(0x0083, iadev->seg_reg+VCT_BASE);  
	writew(0x0060, iadev->seg_reg+VCTE_BASE);  
  
	vc = (struct main_vc *)(iadev->seg_ram+MAIN_VC_TABLE);  
	evc = (struct ext_vc *)(iadev->seg_ram+EXT_VC_TABLE);  
	for(i=0; i<SCHEDSZ; i++)  
	{  
		memset((caddr_t)vc, 0, sizeof(struct main_vc));  
		memset((caddr_t)evc, 0, sizeof(struct ext_vc));  
              	iadev->testTable[i].lastTime = 0;
 		iadev->testTable[i].fract = 0;
                iadev->testTable[i].vc_status = VC_UBR;
		vc++;  
		evc++;  
	}  
  
	/* Other Initialization */  
	  
	/* Max Rate Register */  
	writew(RATE155, iadev->seg_reg+MAXRATE);  
  
	/* Traffic Management Parameters Register */  
	//writew(STPARMS_1K | UBR_EN | ABR_EN, iadev->seg_reg+STPARMS);  

        // pwang_test
	writew((UBR_EN | ABR_EN | (0x23 << 2)), iadev->seg_reg+STPARMS);  
	// writew(ABR_EN, iadev->seg_reg+STPARMS);  
  
	/* Set Idle Header Reigisters to be sure */  
	writew(0, iadev->seg_reg+IDLEHEADHI);  
	writew(0, iadev->seg_reg+IDLEHEADLO);  
  
	/* Program ABR UBR Priority Register  as  PRI_ABR_UBR_EQUAL */
        writew(0xaa00, iadev->seg_reg+ABRUBR_ARB); 

	/* RM Cell Protocol ID and Message Type */  
	writew(RM_TYPE_4_0, iadev->seg_reg+RM_TYPE);  
#if 0
	/* Copy real ffred registers into the local copy */
	for (i=0; i<(sizeof (ffredn_t))/4; i++)
                ((u_int *)ffL)[i] = ((u_int *)iadev->seg_reg)[i] & 0xffff;
#endif
  
	/* Mode Register 1 */  
	writew(MODE_REG_1_VAL, iadev->seg_reg+MODE_REG_1);  
  
	/* Mode Register 0 */  
	writew(T_ONLINE, iadev->seg_reg+MODE_REG_0);  
  
	/* Interrupt Status Register - read to clear */  
	readw(iadev->seg_reg+SEG_INTR_STATUS_REG);  
  
	/* Interrupt Mask Reg- don't mask TCQ_NOT_EMPTY interrupt generation */  
	writew(~TCQ_NOT_EMPTY, iadev->seg_reg+SEG_MASK_REG);  
  
	skb_queue_head_init(&iadev->tx_dma_q);  
  
	return 0;  
}  
  
  
  
  
/* common stuff for interrupt handling etc */  
/*------------------------------------ common ---------------------------------*/  
  
static void ia_int(int irq, void *dev_id, struct pt_regs *regs)  
{  
	struct atm_dev *dev;  
	IADEV *iadev;  
	unsigned int status;  
  
	dev = dev_id;  
	iadev = IPHASE5575_DEV(dev);  
	while( (status = readl(iadev->reg+IPHASE5575_BUS_STATUS_REG)))  
	{  
		if (status & STAT_REASSINT)  
		{  
			/* do something */  
//			DPRINTK("REASSINT Bus status reg: %08x\n", status);  
			rx_intr(dev);  
		}  
		if (status & STAT_SEGINT)  
		{  
			/* do something */  
			tx_intr(dev);  
		}  
		if (status & STAT_DLERINT)  
		{  
			/* do something */  
			/* Clear this bit by writing a 1 to it. */  
			/* completion of receive DLE interrupt  
				- means completion of DMA operation   
				- free the buffers etc   
			*/  
			*(u_int *)(iadev->reg+IPHASE5575_BUS_STATUS_REG) = STAT_DLERINT;  
			DPRINTK("DLE R INT Bus status reg: %08x\n", status);  
			rx_dle_intr(dev);  
		}  
		if (status & STAT_DLETINT)  
		{  
			/* do something */  
			*(u_int *)(iadev->reg+IPHASE5575_BUS_STATUS_REG) = STAT_DLETINT;  
			DPRINTK("ia_int DLE T INT\n");  
			tx_dle_intr(dev);  
		}  
		if (status & (STAT_FEINT | STAT_ERRINT | STAT_MARKINT))  
		{  
			/* do something */  
			DPRINTK("ia_int MISC INT\n");  
		}  
	}  
}  
  
  
  
/*------------------------------------ entries --------------------------------*/  
static int get_esi(struct atm_dev *dev)  
{  
	IADEV *iadev;  
	int i;  
	u32 mac1;  
	u16 mac2;  
  
	iadev = IPHASE5575_DEV(dev);  
	mac1 = cpu_to_be32(le32_to_cpu(readl(  
			iadev->reg+IPHASE5575_MAC1)));  
	mac2 = cpu_to_be16(le16_to_cpu(readl(  
			iadev->reg+IPHASE5575_MAC2)));  
  
	DPRINTK("ESI: 0x%08x%04x\n", mac1, mac2);  
	for (i=0; i<MAC1_LEN; i++)  
		dev->esi[i] = mac1 >>(8*(MAC1_LEN-1-i));  
  
	for (i=0; i<MAC2_LEN; i++)  
		dev->esi[i+MAC1_LEN] = mac2 >>(8*(MAC2_LEN - 1 -i));  
  
	return 0;  
}  
  
static int reset_sar(struct atm_dev *dev)  
{  
        IADEV *iadev;  
        int i, error;  
        unsigned int pci[64];  
  
        iadev = IPHASE5575_DEV(dev);  
        for(i=0; i<64; i++)  
                if ((error = pci_read_config_dword(iadev->pci,  
                                        i*4, &pci[i])) != PCIBIOS_SUCCESSFUL)  
                        return error;  
  
        writel(0, iadev->reg+IPHASE5575_EXT_RESET);  
  
        for(i=0; i<64; i++)  
                if ((error = pci_write_config_dword(iadev->pci,  
                                        i*4, pci[i])) != PCIBIOS_SUCCESSFUL)  
                        return error;  
  
	DPRINTK("Delay for 5 us\n");  
	udelay(5);  
	return 0;  
}  
  
  
__initfunc(static int ia_init(struct atm_dev *dev))  
{  
	IADEV *iadev;  
	unsigned int real_base, base;  
	unsigned short command;  
	unsigned char revision;  
	int error, i;  
  
	/* The device has been identified and registered. Now we read   
	   necessary configuration info like memory base address,   
	   interrupt number etc */  
  
	DPRINTK(">ia_init\n");  
	dev->ci_range.vpi_bits = 0;  
	dev->ci_range.vci_bits = NR_VCI_LD;  
	iadev = IPHASE5575_DEV(dev);  
	  
	if ((error = pci_read_config_word(iadev->pci,   
			PCI_COMMAND,&command))   
	    || (error = pci_read_config_dword(iadev->pci,   
			PCI_BASE_ADDRESS_0,&real_base))   
	    || (error = pci_read_config_byte(iadev->pci,   
			PCI_INTERRUPT_LINE,&iadev->irq))   
	    || (error = pci_read_config_byte(iadev->pci,   
			PCI_REVISION_ID,&revision)))   
	{  
		printk(KERN_ERR DEV_LABEL "(itf %d): init error 0x%x\n",  
                	dev->number,error);  
                return -EINVAL;  
        }  
        DPRINTK(KERN_NOTICE DEV_LABEL "(itf %d): rev.%d,realbase=0x%x,irq=%d\n",  
		dev->number, revision, real_base, iadev->irq);  
  
	/* find mapping size of board */  
  
	/* write all 1's into the base address register.  
	   read the register whic returns us 0's in the don't care bits.  
	   size is calculated as ~(don't cre bits) + 1 */  
  
	if (pci_write_config_dword(iadev->pci,   
			PCI_BASE_ADDRESS_0, 0xffffffff)!=PCIBIOS_SUCCESSFUL)  
        {  
                printk(DEV_LABEL "(itf %d): init error 0x%x\n",dev->number,  
                    error);  
                return -EINVAL;  
        }  
	if(pci_read_config_dword(iadev->pci,   
			PCI_BASE_ADDRESS_0,   
			&(iadev->pci_map_size)) !=PCIBIOS_SUCCESSFUL)   
	{  
                printk(DEV_LABEL "(itf %d): init error 0x%x\n",dev->number,  
                    error);  
                return -EINVAL;  
        }  
	iadev->pci_map_size &= PCI_BASE_ADDRESS_MEM_MASK;  
	iadev->pci_map_size = ~iadev->pci_map_size + 1;  
  
	DPRINTK(DEV_LABEL "map size: %i\n", iadev->pci_map_size);  
        if(pci_write_config_dword(iadev->pci,  
			PCI_BASE_ADDRESS_0,  
			real_base)!=PCIBIOS_SUCCESSFUL)  
        {  
                printk(DEV_LABEL "(itf %d): init error 0x%x\n",dev->number,  
                    error);  
                return -EINVAL;  
        }  
  
  
        real_base &= MEM_VALID; /* strip flags (last 4 bits )  
						---> mask with 0xfffffff0 */  
	/* enabling the responses in memory space */  
	command |= (PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);  
        if ((error = pci_write_config_word(iadev->pci,  
                                               PCI_COMMAND,  
                                               command)))  
        {  
                printk(DEV_LABEL "(itf %d): can't enable memory (0x%x)\n",  
                    dev->number,error);  
                return error;  
        }  
	/*  
	 * Delay at least 1us before doing any mem accesses (how 'bout 10?)  
	 */  
	udelay(10);  
  
	/* mapping the physical address to a virtual address in address space */  
        base=(unsigned long)ioremap((unsigned long)real_base,iadev->pci_map_size);  /* ioremap is not resolved ??? */  
  
	if (!base)  
	{  
                printk(DEV_LABEL "(itf %d): can't set up page mapping\n",  
                    dev->number);  
                return error;  
        }  
        printk(KERN_NOTICE DEV_LABEL "(itf %d): rev.%d,base=0x%x,irq=%d\n",  
		dev->number, revision, base, iadev->irq);  
  
	/* filling the iphase dev structure */  
	iadev->mem = iadev->pci_map_size /2;  
	iadev->base_diff = real_base - base;  
	iadev->real_base = real_base;  
	iadev->base = base;  
	  
	/* Bus Interface Control Registers */  
	iadev->reg = (u32 *) (base + REG_BASE);  
	/* Segmentation Control Registers */  
	iadev->seg_reg = (u32 *) (base + SEG_BASE);  
	/* Reassembly Control Registers */  
	iadev->reass_reg = (u32 *) (base + REASS_BASE);  
	/* Front end/ DMA control registers */  
	iadev->phy = (u32 *) (base + PHY_BASE);  
	iadev->dma = (u32 *) (base + PHY_BASE);  
	/* RAM - Segmentation RAm and Reassembly RAM */  
	iadev->ram = (u32 *) (base + ACTUAL_RAM_BASE);  
	iadev->seg_ram =  (base + ACTUAL_SEG_RAM_BASE);  
	iadev->reass_ram = (base + ACTUAL_REASS_RAM_BASE);  
  
	/* lets print out the above */  
	DPRINTK("Base addrs: %08x %08x %08x \n %08x %08x %08x %08x\n", (u32)iadev->reg,(u32)iadev->seg_reg,(u32)iadev->reass_reg, (u32)iadev->phy, (u32)iadev->ram, (u32)iadev->seg_ram, (u32)iadev->reass_ram);  
	  
	/*test board memory, find amount of memory installed ??? */  
	  
//	memset((void *)iadev->ram, 0, iadev->mem);  
	  
	/* lets try reading the MAC address */  
	error = get_esi(dev);  
	if (error) return error;  
	for (i=0; i < ESI_LEN; i++)  
                printk("%s%02X",i ? "-" : "",dev->esi[i]);  
        printk("\n");  
  
        /* reset SAR */  
	DPRINTK("Iphase5575: resetting SAR\n");  
        reset_sar(dev);  
  
	return 0;  
}  
  
  
__initfunc(static int ia_start(struct atm_dev *dev))  
{  
	IADEV *iadev;  
	int error;  
	unsigned char phy;  
	u32 ctrl_reg;  
  
	DPRINTK(">ia_start\n");  
	iadev = IPHASE5575_DEV(dev);  
        if (request_irq(iadev->irq, &ia_int, SA_SHIRQ, DEV_LABEL, dev)) {  
                printk(KERN_ERR DEV_LABEL "(itf %d): IRQ%d is already in use\n",  
                    dev->number, iadev->irq);  
                return -EAGAIN;  
        }  
        /* @@@ should release IRQ on error */  
	/* enabling memory + master */  
        if ((error = pci_write_config_word(iadev->pci,   
				PCI_COMMAND,   
				PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER )))   
	{  
                printk(KERN_ERR DEV_LABEL "(itf %d): can't enable memory+"  
                    "master (0x%x)\n",dev->number, error);  
                return error;  
        }  
	udelay(10);  
  
	/* Maybe we should reset the front end, initialize Bus Interface Control   
		Registers and see. */  
  
	DPRINTK("Bus ctrl reg: %08x\n", readl(iadev->reg+IPHASE5575_BUS_CONTROL_REG));  
	ctrl_reg = readl(iadev->reg+IPHASE5575_BUS_CONTROL_REG);  
	ctrl_reg = (ctrl_reg & (CTRL_LED | CTRL_FE_RST))  
			| CTRL_B8  
			| CTRL_B16  
			| CTRL_B32  
			| CTRL_B48  
			| CTRL_B64  
			| CTRL_B128  
			| CTRL_ERRMASK  
			| CTRL_DLETMASK		/* shud be removed l8r */  
			| CTRL_DLERMASK  
			| CTRL_SEGMASK  
			| CTRL_REASSMASK 	  
			| CTRL_FEMASK  
			| CTRL_CSPREEMPT;  
  
       writel(ctrl_reg, iadev->reg+IPHASE5575_BUS_CONTROL_REG);   
  
	DPRINTK("Bus ctrl reg after initializing: %08x\n", readl(iadev->reg+IPHASE5575_BUS_CONTROL_REG));  
	DPRINTK("Bus status reg after init: %08x\n", readl(iadev->reg+IPHASE5575_BUS_STATUS_REG));  
  
	/* figure out the base addresses of internal tables  
		- segmentation buffer descriptor table.  
	   Actually, internal segmentation and reassembly registers  
	   have to be programmed for corresponding memory maps. How do  
	   we decide the memory map -   
		a) just fix one up from the appendix examples.  
	      	b) fix one each for each of the memory sizes  
			- in this case how is a memory map decided upon ???  
			- now I know how to do this :)  
		c) fully program stuff - duh!!! how ???  
  
	   1023 VC operation control memory map is chosen. There is one suggested in appendix A. An Alternate memory map suggested in 5575 PCI ATM Driver Porting Guide page 7 is as below. Currently, this is programmed in the driver.  
  
	   Transmit   
	   --------  
  
	 	Buffer descr 	0x0000 (128 - 4K)  
		Commn queues	0x1000	Transmit comp, Packet ready(0x1400)   
					(256 - 1K) each  
		CBR Table 	0x1800 (as needed)  
		UBR Table	0x3000 (1K - 4K)  
		UBR Wait queue	0x4000 (1K - 4K)  
		ABR sched	0x5000	and ABR wait queue (1K - 2K) each   
		extended VC	0x6000 (1K - 8K)  
		VC Table	0x8000 (1K - 32K)  
  
	1023 VC operation control memory map chosen from appendix A for receive side.  
  
	Receive   
	-------  
  
		Buffer descr	0x0000 (736 - 23K)  
		VP Table	0x5c00 (256 - 512)  
		Except q	0x5e00 (128 - 512)  
		Free buffer q	0x6000 (1K - 2K)  
		Packet comp q	0x6800 (1K - 2K)  
		Reass Table	0x7000 (1K - 2K)  
		VC Table	0x7800 (1K - 2K)  
		ABR VC Table	0x8000 (1K - 32K)  
	*/  
  
	/* start transmit and receive and do any other enables needed */  
	/* list of things that probably needs to be done here  
		- allocate memory for Tx and Rx DLEs and put the base  
			addresses in the base addr registers.   
		- initialize a queue of descriptor numbers to be used  
			for packet segmentation  
	*/  
  
	error = tx_init(dev);  
	if (error) return error;  
	error = rx_init(dev);  
	if (error) return error;  
  
	ctrl_reg = readl(iadev->reg+IPHASE5575_BUS_CONTROL_REG);  
       	writel(ctrl_reg | CTRL_FE_RST, iadev->reg+IPHASE5575_BUS_CONTROL_REG);   
	DPRINTK("Bus ctrl reg after initializing: %08x\n", readl(iadev->reg+IPHASE5575_BUS_CONTROL_REG));  
	if ((phy=ia_phy_get(dev,0)) == 0x30)  
		printk("pm5346,rev.%d\n",phy&0x0f);  
	else  
		printk("utopia,rev.%0x\n",phy);  
	error = suni_init(dev);  
	if (error) return error;  
  
  
  
	/* bring suni up */  
	/* This is dangerous !!!  
		sends the suni into a big get and put loop ???  
		and the system crashes while removing the module   
	   Apparently, suni.c still makes calls to the SAR driver, so rmmod   
	   won't work. Need to add deallocation support everywhere.   
	*/  
/*
	error = dev->phy->start(dev);  
	if (error) return error;   
*/
	return 0;  
}  
  
static void ia_close(struct atm_vcc *vcc)  
{  
        u16 *vc_table;
        IADEV *iadev;
        unsigned long flags;

	DPRINTK(">ia_close\n");  
	if (!IPHASE5575_VCC(vcc)) return;  
        atomic_set(&vcc->rx_inuse, 0);
	vcc->flags &= ~ATM_VF_READY;  
        save_flags(flags); cli();
        iadev = IPHASE5575_DEV(vcc->dev);
        vc_table = (u16 *)(iadev->reass_ram+RX_VC_TABLE);
        vc_table += vcc->vci; 
        *vc_table |= 15;      /* for invalid VCI */
        // Drain the packets
        udelay(1);
        rx_dle_intr(vcc->dev); 
        iadev->rx_open[vcc->vci] = 0;
        restore_flags(flags);
	kfree(IPHASE5575_VCC(vcc));  
	IPHASE5575_VCC(vcc) = NULL;  
	vcc->flags &= ~ATM_VF_ADDR;  
        
}  
  
static int ia_open(struct atm_vcc *vcc, short vpi, int vci)  
{  
	IADEV *iadev;  
	struct ia_vcc *ia_vcc;  
	int error;  
  
	if (!(vcc->flags & ATM_VF_PARTIAL))  
	{  
		DPRINTK("iphase open: not partially allocated resources\n");  
		IPHASE5575_VCC(vcc) = NULL;  
	}  
	iadev = IPHASE5575_DEV(vcc->dev);  
	error = atm_find_ci(vcc, &vpi, &vci);  
	if (error)   
	{  
	    DPRINTK("iadev: atm_find_ci returned error %d\n", error);  
	    return error;  
	}  
	vcc->vpi = vpi;  
	vcc->vci = vci;  
	if (vci != ATM_VPI_UNSPEC && vpi != ATM_VCI_UNSPEC)  
	{  
		DPRINTK("iphase open: unspec part\n");  
		vcc->flags |= ATM_VF_ADDR;  
	}  
	if (vcc->qos.aal != ATM_AAL5)  
		return -EINVAL;  
	DPRINTK(DEV_LABEL "(itf %d): open %d.%d\n", vcc->dev->number, vcc->vpi, vcc->vci);  
  
	/* Device dependent initialization */  
	ia_vcc = kmalloc(sizeof(struct ia_vcc), GFP_KERNEL);  
	if (!ia_vcc) return -ENOMEM;  
	IPHASE5575_VCC(vcc) = ia_vcc;  
  
	if ((error = open_rx(vcc)))  
	{  
		DPRINTK("iadev: error in open_rx, closing\n");  
		ia_close(vcc);  
		return error;  
	}  
  
	if ((error = open_tx(vcc)))  
	{  
		DPRINTK("iadev: error in open_tx, closing\n");  
		ia_close(vcc);  
		return error;  
	}  
  
	vcc->flags |= ATM_VF_READY;  
	DPRINTK("ia open returning\n");  
	return 0;  
}  
  
static int ia_change_qos(struct atm_vcc *vcc, struct atm_qos *qos, int flags)  
{  
	DPRINTK(">ia_change_qos\n");  
  
	return 0;  
}  
  
static int ia_ioctl(struct atm_dev *dev, unsigned int cmd, void *arg)  
{  
   PIA_CMDBUF ia_cmds;
   IADEV *iadev = IPHASE5575_DEV(dev);
   int i;
   u16 *tmps;
   DPRINTK(">ia_ioctl\n");  
   if (cmd != IA_CMD) return 0;
   ia_cmds	= (PIA_CMDBUF)arg;
   switch (ia_cmds->cmd) {
   case MEMDUMP:
   {
	switch (ia_cmds->sub_cmd) {
       	  case MEMDUMP_DEV:     
             memcpy((char*)ia_cmds->buf, (char*)iadev,
                          			sizeof(IADEV));
             ia_cmds->status = 0;
             break;
          case MEMDUMP_SEGREG:
             tmps = (u16 *)ia_cmds->buf;
             for(i=0; i<0x80; i+=2, tmps++)
                *tmps = *(u16*)(iadev->seg_reg+i);
             ia_cmds->status = 0;
             ia_cmds->len = 0x80;
             break;
          case MEMDUMP_REASSREG:
             tmps = (u16 *)ia_cmds->buf;
             for(i=0; i<0x80; i+=2, tmps++)
                *tmps = *(u16*)(iadev->reass_reg+i);
             ia_cmds->status = 0;
             ia_cmds->len = 0x80;
             break;
          case MEMDUMP_FFL:
          {  
             ia_regs_t       regs_local;
             ffredn_t        *ffL = &regs_local.ffredn;
             rfredn_t        *rfL = &regs_local.rfredn;
                     
             /* Copy real rfred registers into the local copy */
 	     for (i=0; i<(sizeof (rfredn_t))/4; i++)
                ((u_int *)rfL)[i] = ((u_int *)iadev->reass_reg)[i] & 0xffff;
             	/* Copy real ffred registers into the local copy */
	     for (i=0; i<(sizeof (ffredn_t))/4; i++)
                ((u_int *)ffL)[i] = ((u_int *)iadev->seg_reg)[i] & 0xffff;

             memcpy((char*)ia_cmds->buf,(char*)&regs_local,sizeof(ia_regs_t));
             ia_cmds->status = 0;                  
	 }	
    	     break;        
         case READ_REG:
         {  
             desc_dbg(iadev); 
             ia_cmds->status = 0; 
         }
             break;
         default:
             memcpy((char*)ia_cmds->buf, (char*)ia_cmds->maddr, ia_cmds->len);
             ia_cmds->status = 0;
             break;
      }	
   }
      break;
   default:
      break;

   }	
   return 0;  
}  
  
static int ia_getsockopt(struct atm_vcc *vcc, int level, int optname,   
	void *optval, int optlen)  
{  
	DPRINTK(">ia_getsockopt\n");  
  
	return -EINVAL;  
}  
  
static int ia_setsockopt(struct atm_vcc *vcc, int level, int optname,   
	void *optval, int optlen)  
{  
	DPRINTK(">ia_setsockopt\n");  
  
	return -EINVAL;  
}  
  
static int ia_send(struct atm_vcc *vcc, struct sk_buff *skb)  
{  
	IADEV *iadev;  
	struct dle *wr_ptr;  
	struct tx_buf_desc *buf_desc_ptr;  
	int desc;  
	int comp_code;  
	unsigned int addr;  
	int total_len, pad, last;  
	struct cpcs_trailer *trailer;  
//	ffredn_t        *ffL = &brd_regs_local.ffredn; 
        unsigned long flags;
 
        save_flags(flags); cli(); 
        
	iadev = IPHASE5575_DEV(vcc->dev);  
  
	if (!skb)  
	{  
              	printk(KERN_CRIT "null skb in ia_send\n");  
               	dev_kfree_skb(skb);  
                restore_flags(flags);
               	return -EINVAL;  
        }  
  
        ATM_SKB(skb)->vcc = vcc;
  
	/* Get a descriptor number from our free descriptor queue  
	   We get the descr number from the TCQ now, since I am using  
	   the TCQ as a free buffer queue. Initially TCQ will be   
	   initialized with all the descriptors and is hence, full.  
	*/
	desc = get_desc (iadev);
	if (desc == 0xffff) {
	    // TODO: update tcq_empty counter
            printk(DEV_LABEL " No desc available!!\n");
            desc_dbg(iadev);
 	    restore_flags(flags);
	    return -EINVAL;
        }      

	comp_code = desc >> 13;  
	desc &= 0x1fff;  
  
	if ((desc == 0) || (desc > TX_BUFFERS))  
	{  
		printk(DEV_LABEL "invalid desc for send: %d\n", desc); 
                vcc->stats->tx++;
		if (vcc->pop)   
		    vcc->pop(vcc, skb);   
		else  
		    dev_kfree_skb(skb);
 		restore_flags(flags);
		return 0;   /* return SUCCESS */
	}  
	DPRINTK("ia /end desc: %d\n", desc);  
  
	if (comp_code)  
	{  
	    printk(DEV_LABEL "send desc:%d completion code %d error\n", desc, comp_code);  
	}  
       
        iadev->ffL.tcq_rd += 2;
        if (iadev->ffL.tcq_rd > iadev->ffL.tcq_ed)
	  	iadev->ffL.tcq_rd  = iadev->ffL.tcq_st;
	writew(iadev->ffL.tcq_rd, iadev->seg_reg+TCQ_RD_PTR);
  
	/* Put the descriptor number in the packet ready queue  
		and put the updated write pointer in the DLE field   
	*/   
	*(u16*)(iadev->seg_ram+iadev->ffL.prq_wr) = desc; 

 	iadev->ffL.prq_wr += 2;
        if (iadev->ffL.prq_wr > iadev->ffL.prq_ed)
                iadev->ffL.prq_wr = iadev->ffL.prq_st;
	  
	/* Figure out the exact length of the packet and padding required to make it  
		aligned on a 48 byte boundary.  
	*/  
	total_len = skb->len + sizeof(struct cpcs_trailer);  
	last = total_len - (total_len/48)*48;  
	pad = 48 - last;  
	total_len = pad + total_len;  
	DPRINTK("ia packet len:%d padding:%d\n", total_len, pad);  
 
	/* Put the packet in a tx buffer */   
	if (!iadev->tx_buf[desc-1])  
		printk("couldn't get free page\n");  
	memset(iadev->tx_buf[desc-1], 0, total_len);  
	memcpy(iadev->tx_buf[desc-1], skb->data, skb->len);  
	addr = virt_to_bus(iadev->tx_buf[desc-1]);  

	/* Build the CS trailer */  
	trailer = (struct cpcs_trailer*)(iadev->tx_buf[desc-1] + total_len - sizeof(struct cpcs_trailer));  
	trailer->control = 0;  
	trailer->length = ((skb->len & 0xff) << 8) | ((skb->len & 0xff00) >> 8); /*big endian*/  
	trailer->crc32 = 0;	/* not needed - dummy bytes */  

#ifdef DEBUG_TXPKT
	/* Display the packet */  
	//printk("Sent data: len = %d MsgNum = %d\n", skb->len, tcnter++);  
        //xdump(skb->data, skb->len, "TX: ");
        //printk("\n");
#endif /* DEBUG_TXPKT */

	/* Build the buffer descriptor */  
	buf_desc_ptr = (struct tx_buf_desc *)(iadev->seg_ram+TX_DESC_BASE);  
	buf_desc_ptr += desc;	/* points to the corresponding entry */  
	buf_desc_ptr->desc_mode = AAL5 | EOM_EN | APP_CRC32;	/* shud have worked with just   
							the auto CRC append enabled in the VC  
							Tbl entry but doesn't ??? */  
	buf_desc_ptr->vc_index = vcc->vci;	/* Now I am getting doubtful if this is right. */  
	buf_desc_ptr->bytes = total_len;  
  
	clear_lockup (vcc, iadev);

	/* Build the DLE structure */  
	wr_ptr = iadev->tx_dle_q.write;  
	memset((caddr_t)wr_ptr, 0, sizeof(struct dle));  
	wr_ptr->sys_pkt_addr = addr;  
	wr_ptr->local_pkt_addr = (buf_desc_ptr->buf_start_hi << 16) | buf_desc_ptr->buf_start_lo;  
	/* wr_ptr->bytes = swap(total_len);	didn't seem to affect ?? */  
	wr_ptr->bytes = total_len;  
	wr_ptr->mode = 0; /* IA_ABR fix pwang_test DMA_INT_ENABLE; */
	wr_ptr->prq_wr_ptr_data = iadev->ffL.prq_wr;
  
	/* end is not to be used for the DLE q */  
	if (++wr_ptr == iadev->tx_dle_q.end)  
		wr_ptr = iadev->tx_dle_q.start;  
  
	iadev->tx_dle_q.write = wr_ptr;  
	if (vcc->pop)  
	{  
		DPRINTK("protocol pop\n");  
		vcc->pop(vcc, skb);  
	}  
	else  
		dev_kfree_skb(skb);  
        vcc->stats->tx++;

	/* Increment transaction counter */  
	writel(1, iadev->dma+IPHASE5575_TX_COUNTER);  
	DPRINTK("ia send done\n");  
        restore_flags(flags);
	return 0;  
}  
  
static int ia_sg_send(struct atm_vcc *vcc, unsigned long start,   
	unsigned long size)  
{  
	DPRINTK(">ia_sg_send\n");  
  
	return 0;  
}  
  
static void ia_phy_put(struct atm_dev *dev, unsigned char value,   
	unsigned long addr)  
{  
	//DPRINTK(">ia_phy_put\n");  
	writel(value,IPHASE5575_DEV(dev)->phy+addr);  
}  
  
static unsigned char ia_phy_get(struct atm_dev *dev, unsigned long addr)  
{  
	//DPRINTK(">ia_phy_get\n");  
	return readl(IPHASE5575_DEV(dev)->phy+addr);  
}  
  
  
static const struct atmdev_ops ops = {  
	NULL,			/* no dev close */  
	ia_open,  
	ia_close,  
	ia_ioctl,  
	ia_getsockopt,  
	ia_setsockopt,  
	ia_send,  
	ia_sg_send,  
	NULL,			/* no send_oam */  
	ia_phy_put,  
	ia_phy_get,  
	NULL,			/* no feedback */  
	ia_change_qos,  
	NULL,  
	NULL,  
};  
	  
  
__initfunc(int ia_detect(void))  
{  
	struct atm_dev *dev;  
	IADEV *iadev;  
	struct pci_dev *prev_dev;                                                
	int index;  
  
	DPRINTK(">ia_detect\n");  
	if (!pci_present()) {  
		printk(KERN_ERR DEV_LABEL " driver but no PCI BIOS ?\n");  
		return 0;  
	}  
	iadev = (IADEV *)kmalloc(sizeof(  
		IADEV), GFP_KERNEL);  
	if (!iadev) return -ENOMEM;  
	index = 0;  
	prev_dev = NULL;  
	while((iadev->pci = pci_find_device(PCI_VENDOR_ID_IPHASE,  
		PCI_DEVICE_ID_IPHASE_5575,  prev_dev))) {  
		DPRINTK("ia detected at bus:%d device: %d function:%d\n" ,iadev->pci->bus->number, PCI_SLOT(iadev->pci->devfn), PCI_FUNC(iadev->pci->devfn));  
		dev = atm_dev_register(DEV_LABEL, &ops, -1, 0);   
			/* -1 --- is for number  
			    0 --- flags  
			    don't know what it means yet.   
			*/  
		if (!dev) break;  
		DPRINTK(DEV_LABEL "registered at (itf :%d)\n", dev->number);  
		IPHASE5575_DEV(dev) = iadev; 
#ifdef IA_ABR 
                // TODO: multi_board using ia_boards logic in cleanup_module
		ia_dev = iadev;
#endif 
		if (ia_init(dev) || ia_start(dev)) {  
			atm_dev_deregister(dev);  
			break;  
		}  
		prev_dev = iadev->pci;  
		iadev->next_board = ia_boards;  
		ia_boards = dev;  
		iadev = (IADEV *)kmalloc(   
				sizeof(IADEV), GFP_KERNEL);  
		if (!iadev) break;   
		index++;  
	}  
	printk("end of iphase_detect\n");  
	return index;  
}  
  

static void ia_led_timer(unsigned long arg) {
 	unsigned long flags;
  	static u_char blinking = 0;
        static u32 ctrl_reg; 
	save_flags(flags); cli();
	ctrl_reg = readl(ia_dev->reg+IPHASE5575_BUS_CONTROL_REG);
	if (blinking == 0) {
		blinking++;
                ctrl_reg &= (~CTRL_LED);
                writel(ctrl_reg, ia_dev->reg+IPHASE5575_BUS_CONTROL_REG);

        }
        else {
		blinking = 0;
		ctrl_reg |= CTRL_LED;
                writel(ctrl_reg, ia_dev->reg+IPHASE5575_BUS_CONTROL_REG);
        }
	mod_timer(&ia_timer, jiffies + HZ / 4);
        restore_flags(flags);
 	return;
}

#ifdef MODULE  
  
int init_module(void)  
{  
       printk("<1>Enter init_module\n");
	DPRINTK(">ia init_module\n");  
	if (!ia_detect()) {  
		printk(KERN_ERR DEV_LABEL ": no adapter found\n");  
		return -ENXIO;  
	}  
	// MOD_INC_USE_COUNT; 
        ia_timer.next = NULL;
       	ia_timer.prev = NULL;
   	ia_timer.expires = jiffies + 3*HZ;
   	ia_timer.data = 0UL;
   	ia_timer.function = ia_led_timer;
   	add_timer(&ia_timer); 
   
	return 0;  
}  
  
  
void cleanup_module(void)  
{  
	struct atm_dev *dev;  
	IADEV *iadev;  
	unsigned short command;  
        int i;
 
	DPRINTK(">ia cleanup_module\n");  
	/* They don't get rid of atm drivers, don't know why ???  
	   so nothing here yet */  
	/* But lets anyway try */  
        // MOD_DEC_USE_COUNT;
 
	if (MOD_IN_USE)  
		printk("ia: module in use\n");  
        del_timer(&ia_timer);
	while(ia_boards)  
	{  
		dev = ia_boards;  
		iadev = IPHASE5575_DEV(dev);  
		ia_boards = iadev->next_board;  
      		/* De-register device */  
      		atm_dev_deregister(dev);  
		DPRINTK(">iadev deregistered at (itf:%d)\n", dev->number);  
		for (i= 0; i< TX_BUFFERS; i++)
                        kfree(iadev->tx_buf[i]);
      		/* Disable memory mapping and busmastering */  
		if (pci_read_config_word(iadev->pci,  
					     PCI_COMMAND, &command) != 0)  
      		{  
         		printk("ia: can't read PCI_COMMAND.\n");  
      		}  
      		command &= ~(PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);  
      		if (pci_write_config_word(iadev->pci,  
					      PCI_COMMAND, command) != 0)  
      		{  
         		printk("ia: can't write PCI_COMMAND.\n");  
      		}  
      		free_irq(iadev->irq, dev);  
      		iounmap((void *) iadev->base);  
      		kfree(iadev);  
	}  
	/* and voila whatever we tried seems to work. I don't know if it will  
		fix suni errors though. Really doubt that. */  
  
}  
  
#endif  
  


